import sql
from fields import *
from manager import Manager
from categories import *


model_cache = {}

class ModelBase(type):
    def __new__(cls, cls_name, bases, attrs):
        # Check if this is a subclass of Model.
        try:
            if not [b for b in bases if issubclass(b, Model)]:
                return super(ModelBase, cls).__new__(cls, cls_name, bases, attrs)
        except NameError:
            return super(ModelBase, cls).__new__(cls, cls_name, bases, attrs)

        # See if we've already defined a model with this name.
        if cls_name in Model.model_cache:
            # Return the created class.
            return Model.model_cache[cls_name]

        newclass = type.__new__(cls, cls_name, bases, attrs)
        newclass._fields = []
        newclass._field_names = {}

        # Add manager to class
        newclass.objects = Manager(newclass)

        # Add meta to class
        newclass._meta = Options(newclass)
        newclass._meta.attrs = attrs

        # Add id field
        af = AutoField(default=None, null=True)
        af.contribute_to_class(newclass, 'id')

        #Check to see if direct descendant of Model
        if Model in newclass.__bases__:
            #If so, give it a category table.
            newclass.categories = Categories(newclass)

        cf = CategoryField(default=None, null=True)
        cf.contribute_to_class(newclass, 'db_category_id')

        # Process base classes.
        for cls in bases:
            if cls != Model:
                for name in cls._meta.field_order:
                    field = cls._meta.fields[name]

                    if field.inherit and hasattr(field, 'contribute_to_class'):
                        field.contribute_to_class(newclass, name)

        # Add attributes to class.
        fields = []
        for attr, obj in attrs.items():
            if hasattr(obj, 'contribute_to_class'):
                fields.append((attr, obj))
            else:
                setattr(newclass, attr, obj)

        for name, field in sorted(fields, lambda a, b: a[1].creation_counter - b[1].creation_counter):
            field.contribute_to_class(newclass, name)

        Model.model_cache[cls_name] = newclass

        return newclass



class Model(object):
    __metaclass__ = ModelBase
    categories = None

    model_cache = {}

    def __init__(self, *args, **kwargs):
        iter_fields = iter(self._meta.fields.items())

        if args:
            for value, field in zip(args, iter_fields):
                setattr(self, field[0], value)

        for name, field in iter_fields:
            if kwargs:
                value = kwargs.pop(name, field.get_default())

            else:
                value = field.get_default()

            setattr(self, name, value)

    def __setattr__(self, attr, value):
        if attr in self._meta.field_order:
            value = self._meta.fields[attr].to_python(value)

        self.__dict__[attr] = value

    def __repr__(self):
        return u'<%s: %s>' % (self.__class__.__name__, self)

    def __str__(self):
        if hasattr(self, '__unicode__'):
            v = unicode(self).encode('utf-8')
            if v:
                return v
        return '%s object' % self.__class__.__name__

    @classmethod
    def add_field(cls, name, field):
        if name not in cls._meta.field_order:
            cls._meta.field_order.append(name)

            cls._meta.fields[name] = field

    @classmethod
    def sync(cls):
        """Sync model structure to database."""
        cursor = sql.connection.cursor()

        field_names = []
        fields = []
        for name in cls._meta.field_order:
            field = cls._meta.fields[name]

            field_names.append(name)

            if field.primary:
                s = '%s %s PRIMARY KEY'
            else:
                s = '%s %s'

            fields.append(s % (field.db_name, field.db_type))

        field_sql = ', '.join(fields)
        create_table = lambda: sql.execute('CREATE TABLE %s (%s)' % (table, field_sql))



        table = cls._meta.db_table
        # Check if table exists.
        table_info = sql.query('pragma table_info(%s)'% table)
        if table_info:
            # Get existing columns from table.
            cur_columns = [str(c[1]) for c in table_info]

            # Match up columns between old and new tables.
            new_names = []
            old_names = []
            for name in field_names:
                new_names.append(name)

                if name in cur_columns:
                    old_names.append(name)

                else:
                    value = u"'%s'" % cls._meta.fields[name].get_default()

                    old_names.append(value)

            # Create temp table.
            if sql.query('pragma table_info(sync_temp)'):
                sql.execute('DROP TABLE sync_temp')
            sql.execute('CREATE TABLE sync_temp (%s)' % field_sql)

            # Copy existing data into temp table.
            field_names = ', '.join(new_names)
            old_names = ', '.join(old_names)
            sql.execute('INSERT INTO sync_temp (%s) SELECT %s FROM %s' % (field_names, old_names, table))

            # Drop old table.
            sql.execute('DROP TABLE %s' % table)

            # Rename table.
            sql.execute('ALTER TABLE sync_temp RENAME TO %s' % table)

        else:
            # Create new table.
            create_table()


    def save(self):
        self.savebase()

    def savebase(self):
        table = self._meta.db_table

        field_names = []
        field_values = []
        for name, field in self._meta.fields.items():
            value = getattr(self, name)
            if value is None:
                # Check if this field allows NULL.
                if not field.null:
                    raise Exception(u'%s field does not allow None value.' % field.name)

            else:
                value = field.to_db(value)

            field_names.append(field.db_name)
            field_values.append(value)

        # Insert into DB.
        if not self.id:
            field_names = ', '.join(field_names)
            field_sql = ', '.join('?' * len(field_values))

            sql.execute('INSERT INTO %s (%s) VALUES (%s)' % (table, field_names, field_sql), *field_values)
            self.id = sql.last_insert_id()

        # Update existing record.
        else:
            fields = []
            for name, value in zip(field_names, field_values):
                fields.append('%s = ?' % name)

            field_sql = ', '.join(fields)
            sql.execute('UPDATE %s SET %s WHERE id = %s' % (table, field_sql, self.id), *field_values)

    def delete(self):
        if not self.id:
            raise Exception(u'%s object cannot be deleted because its id attribute is set to None.' % self._meta.object_name)

        sql.execute('DELETE FROM %s WHERE id = %s' % (self._meta.db_table, self.id))

    @classmethod
    def is_direct_descendant(cls, parent):
        if parent in cls.__bases__:
            return True

        return False

class Options(object):
    def __init__(self, cls):
        name = cls.__name__

        self.db_table = name.lower()
        self.object_name = cls.__name__
        self.fields = {}
        self.field_order = []
